(** Entry widgets. *)

open Misc
open Tsdl
open Tsdl_ttf
open Widget

[@@@landmark "auto"]

module B = Textbuffer

(** Property used to define the character displayed instead of
  newline in single-line edition widgets. Default value is ↵ (U+21B5 (8629)).*)
let newline_char_codepoint = Props.int_prop
  ~after:[Render]
  ~default:8629
  ~inherits:true "newline_char_codepoint"

class entry ?(class_="entry") ?name ?props () =
  object(self)
    inherit Textview.textview ~class_ ?name ?props () as super
    (**/**)
    val mutable replacing_text = false
    (**/**)

    method newline_char_codepoint = self#get_p newline_char_codepoint
    method set_newline_char_codepoint = self#set_p newline_char_codepoint

    (**/**)
    method! private on_buffer_change change =
      super#on_buffer_change change ;
      if not replacing_text then self#set_p Props.text (self#text())
    (**/**)

    method set_text str =
      replacing_text <- true ;
      (try ignore(self#delete ()) ; replacing_text <- false
      with _ -> replacing_text <- false);
      self#insert str

    (**/**)
    method! render_cursor rend ~offset rg cursor_id c =
      if self#has_focus then
        super#render_cursor rend ~offset rg cursor_id c

    initializer
      Textview.set_default_key_bindings self#as_textview;
      B.set_map_in buffer
        (Some (fun c ->
           match Uchar.to_int c with
           | 10 (* \n *) -> Uchar.of_int self#newline_char_codepoint
           | _ -> c)) ;
      B.set_map_out buffer
        (Some (fun c ->
           if Uchar.to_int c = self#newline_char_codepoint then
             Uchar.of_int 10
           else
             c))

  end


(** Convenient function to create a {!class-entry}. Optional arguments:
  {ul
  {- [editable] specifies whether the contents is editable by user (default is [true]).}
  {- [text] specifies the initial text (default is empty).}
  }
  See {!Widget.widget_arguments} for other arguments. *)
let entry ?class_ ?name ?props ?editable ?text ?pack () =
  let e = new entry ?class_ ?name ?props () in
  Option.iter e#set_editable editable ;
  Widget.may_pack ?pack e#coerce ;
  Option.iter e#insert text ;
  e


(*
class entry ?(class_="entry") ?name ?props () =
  object(self)
    inherit Text.label ~class_ ?name ?props () as super

    (* position of curor. n means there are n characters before. *)
    val mutable cursor_pos = 0
    val mutable text = ""

    method set_editable = self#set_p Props.editable
    method editable = self#get_p Props.editable

    method cursor_pos_from_mouse ~x ~y =
      let x = x - g_inner.x and y = y - g_inner.y in
      let gtext = super#gtext in
      let x = x - gtext.x in
      let y = y - gtext.y in
      match G.inside ~x ~y g_inner with
      | false -> None
      | true ->
          let font = Props.get_font props in
          Some (Text.pos_from_x font text x)

    method render_me ~layer rend ~offset rg =
      super#render_me ~layer rend ~offset rg ;
      match self#opt_p Props.has_focus with
      | Some true -> self#render_cursor rend ~offset rg
      | _ -> ()

    method render_cursor rend ~offset:(x,y) geom =
      let gtext = super#gtext in
      let cx =
        match ttext#font with
        | None -> 0
        | Some font ->
            let t = self#get_p Props.text in
            let len = Utf8.length t in
            let sub = Utf8.sub t ~pos:0 ~len:(min cursor_pos len) in
            (*Log.info (fun m -> m "%s#render_cursor t=%S, len=%d, cursor_pos=%d"
               self#me t len cursor_pos);*)
            let> (w,_) = Ttf.size_utf8 font sub in
            gtext.x + w
      in
      Log.info (fun m -> m "%s#render_cursor x=%d g=%a g_inner=%a"
         self#me x G.pp g G.pp g_inner);
      let h = self#g_inner.h - 2 in
      let w = 2 in
      let rect = G.{ x = cx ; y = 1 ; w ; h } in
      let rect = G.translate ~x:(g.x+g_inner.x) ~y:(g.y+g_inner.y) rect in
      match G.inter rect geom with
      | None -> ()
      | Some rect ->
          let rect = G.translate ~x ~y rect in
          Log.info (fun m -> m "%s#render_cursor rect=%a" self#me G.pp rect);
          Render.fill_rect rend (Some rect) (self#fg_color_now)

    method set_text str =
      let str = Utf8.normalize str in
      cursor_pos <- Utf8.length str ;
      text <- str ;
      super#set_text str

    method insert str =
      (*Log.info (fun m -> m "insert %S with len=%d" str (Utf8.length str));*)
      text <- Utf8.(insert text cursor_pos (normalize str));
      cursor_pos <- cursor_pos + (Utf8.length str);
      super#set_text text

    method get_subtext ~pos ~len =
      let l = Utf8.length text in
      if pos < 0 || pos + len > l then
          None
        else
          Some (Utf8.sub text ~pos ~len)

    method delete_range ~pos ~len =
      let l = Utf8.length text in
      let s1 = self#get_subtext ~pos:0 ~len:pos in
      let s2 = self#get_subtext ~pos:(pos+len) ~len:(l-pos-len) in
      Log.debug (fun m -> m "%s#delete_range text=%S, pos=%d, len=%d"
         self#me text pos len);
      let upd_cursor () =
        if cursor_pos <= pos then
          ()
        else
          if cursor_pos < pos + len then
            cursor_pos <- max 0 pos
          else
            cursor_pos <- max 0 (cursor_pos - len)
      in
      match s1, s2 with
      | None, None -> ()
      | Some s, None
      | None, Some s ->
          text <- s;
          upd_cursor ();
          super#set_text s
      | Some s1, Some s2 ->
          text <- Utf8.concat s1 s2 ;
          upd_cursor ();
          super#set_text text

    method delete_backward =
      match cursor_pos with
      | n when n <= 0 -> ()
      | n -> self#delete_range ~pos:(cursor_pos - 1) ~len:1

    method delete_forward =
      match cursor_pos with
      | n when n >= Utf8.length text -> ()
      | n -> self#delete_range ~pos:cursor_pos ~len:1

    method on_key_down pos ev key mods =
      let need_render =
        match Sdl.Event.(get ev keyboard_keycode) with
        | k when k = Sdl.K.home ->
            cursor_pos <- 0;
            true
        | k when k = Sdl.K.kend ->
            cursor_pos <- Utf8.length text;
            true
        | k when k = Sdl.K.left ->
            let c = cursor_pos in
            cursor_pos <- max 0 (c - 1);
            c <> cursor_pos
        | k when k = Sdl.K.right ->
            let c = cursor_pos in
            cursor_pos <- min (Utf8.length text) (c+1) ;
            c <> cursor_pos
        | k when k = Sdl.K.backspace ->
            self#delete_backward ;
            true
        | k when k = Sdl.K.delete ->
            self#delete_forward ;
            true
        | _ -> super#on_key_down pos ev key mods
      in
      if need_render then
        self#need_render ~layer:(self#get_p Props.layer) g;
      need_render

    method on_sdl_event pos ev =
      match Sdl.Event.(enum (get ev typ)) with
      | `Text_input ->
          let str = Sdl.Event.(get ev text_input_text) in
          (*Log.info (fun m -> m "%s text_input %S" self#me str);*)
          self#insert str ;
          true
      | _ -> super#on_sdl_event pos ev

    method bg_color_now =
      Log.debug (fun m -> m "%s#bg_color_now" self#me);
      self#get_p Props.input_bg_color

    initializer
      Props.(set props focusable true);
      let _ = self#connect Widget.Button_pressed
        (fun ev ->
           let x = ev.x - g.x in
           let y = ev.y - g.y in
           match self#cursor_pos_from_mouse ~x ~y with
           | None -> false
           | Some p ->
               cursor_pos <- p;
               self#need_render ~layer:(self#get_p Props.layer) g;
               false)
      in
      ()
  end

let entry ?class_ ?name ?props ?editable ?text ?pack () =
  let e = new entry ?class_ ?name ?props () in
  Option.iter e#set_editable editable ;
  Widget.may_pack ?pack e#coerce ;
  Option.iter e#set_text text ;
  e
*)
