(** Fonts.

This module provides functions to use fonts through {!Tsdl_ttf.Ttf}.

{!Tsdl_ttf.Ttf.font}s are accessed through font descriptions {!type-font_desc}.
A caching mecanism avoids loading a SDL font each time a it is required.

*)

open Tsdl_ttf

(** Default font size (set to 12 by default). *)
val default_font_size : int ref

(** Font metrics, to gather information from {!Tsdl_ttf}. *)
type font_metrics = {
    font_height : int;
    font_ascent : int;
    font_descent : int;
    font_line_skip : int;
    font_is_fixed_width : int;
}

(** Get all font metrics from the given {!Tsdl_ttf.Ttf.font}. *)
val font_metrics : Ttf.font -> font_metrics

(** A font description describes a font: family, size, style, ... *)
type font_desc = {
    size : int;
    italic : bool;
    bold : bool;
    family : string;
    underline : bool;
    strikethrough : bool;
    kerning : bool;
    outline : int;

}

(** A default font description. *)
val default_font_desc : font_desc

val font_desc_wrapper : font_desc Ocf.Wrapper.t
val font_desc_compare : font_desc -> font_desc -> int

(** Convenient function to create a {!type-font_desc}. Default value for
  optional arguments are:
  {ul
  {- [size]: [!default_font_size],}
  {- [italic]: [false],}
  {- [bold]: [false],}
  {- [strikethrough]: [false],}
  {- [kerning]: [true],}
  {- [outline]: [0].}
  }
*)
val font_desc :
  ?size:int ->
  ?italic:bool ->
  ?bold:bool ->
  ?underline:bool ->
  ?strikethrough:bool ->
  ?kerning:bool ->
  ?outline:int ->
  string ->
  font_desc

val pp_font_desc : Stdlib.Format.formatter -> font_desc -> unit

(** Get a fallback font family associated to given unicode codepoint.*)
val fallback_font : int -> string option

(** [add_fallback_font start stop family] adds a font [family]
  to use as fallback font for the given range [(start..stop)] of unicode
  codepoints.*)
val add_fallback_font : int -> int -> string -> unit

(** Font file extensions. Use when looking for font files in directories.
  Default is [[".ttf"]].*)
val font_exts : string list ref

(** Font directories where to look for fonts, in the form
  [(directory, rec-flag)] where [rec-flag] indicates whether this directory
  should be inspected recursively when looking for fonts.
  {!App.init} will call {!val-load_fonts} which uses [font_dirs] to look
  for available fonts,
  so additional directories must be set before initializing application.
  Default value is [[ Filename.current_dir_name, false ; "/usr/share/fonts/truetype", true ]].
*)
val font_dirs : (string * bool) list ref

(** [load_fonts ()] look for available fonts. Options arguments are:
  {ul
   {- [size]: size used to load a font; default is {!default_font_size}.}
   {- [dirs]: directories to inspect. Default is {!font_dirs}.}
  }
*)
val load_fonts : ?size:int -> ?dirs:(string * bool) list -> unit -> unit Lwt.t

(** Short string for given font desc. *)
val string_of_font_desc : font_desc -> string

(** Close all open font files. A font file remains open because of our caching
  system. If a font file cannot be opened because of two many open files,
  font files will be closed. But in case where a lot of fonts are used and
  the application must open a lot of files, this function may be useful to
  close font files before the application opens files. *)
val close_fonts : unit -> unit

(** Get a SDL font from a font description. *)
val get : font_desc -> Tsdl_ttf.Ttf.font

(** Available fonts. The available fonts are added by calls to {!load_fonts}. *)
val fonts : unit -> font_desc list
