open Stk
open Stk.Misc

let () = Logs.Src.set_level Log.log_src (Some Logs.Warning)
let () = Logs.set_reporter (Misc.lwt_reporter ())

let _ = Tsdl.Sdl.(init Init.(video+events))
let () = Lwt_main.run(App.init())
let canvas_width = 800
let words = 10_000

let is_dir filename =
  match%lwt Lwt_unix.stat filename with
  | exception _ -> Lwt.return_false
  | Unix.{ st_kind = S_DIR } -> Lwt.return_true
  | _ -> Lwt.return_false

type entry =
  { name:string ;
    is_dir: bool ;
    mutable selected: bool;
  }

let mk_win () =
  let (win,scr) = App.create_scrolled_window
    ~flags:Tsdl.Sdl.Window.resizable
    ~w: 400 ~h: 600 "stk tree" in
  let spec =
    let is_leaf e = not e.is_dir in
    let subs e =
      Stk.Log.app (fun m -> m "Listing files from %s" e.name);
      match%lwt Lwt_stream.to_list (Lwt_unix.files_of_directory e.name) with
      | exception _ -> Lwt.return []
      | l ->
          let l = List.filter
            (fun s -> s <> Filename.current_dir_name &&
               s <> Filename.parent_dir_name) l
          in
          Lwt_list.map_s
           (fun s ->
              let name = Filename.concat e.name s in
              let%lwt is_dir = is_dir name in
              Lwt.return
                 { name ; is_dir ; selected = false}
            )
            l
    in
    Stk.Tree.tree_spec ~is_leaf ~subs ()
  in
  let label =
    let create (group:Canvas.group) =
      let t = Canvas.label ~group "" in
      t#set_p Props.focusable true ;
      t#set_p Props.can_focus true ;
      t
    in
    let update ~selected ~expanded e lab =
      let s = Printf.sprintf "%s%s"
        (Filename.basename e.name) (if selected then " (selected)" else "")
      in
      lab#set_text s
    in
    let remove group (lab:Canvas.label) =
      Stk.Log.warn (fun m -> m "removing label with text %S" lab#text);
      group#remove_item (lab#as_full_item)
    in
    Tree.node_label ~create ~update ~remove
  in
  let tree = Stk.Tree.tree ~pack:scr#set_child
    (*~selection_mode:Props.Sel_browse*)
      spec label
  in
  let _ = tree#connect_node_unselected
    (fun (e, node) ->
       Stk.Log.warn (fun m -> m "Unselected %S" e.name);
       false)
  in
  let _ = tree#connect_node_clicked
    (fun (ev, e, node) ->
       match ev.Widget.button with
       | 1 -> false
       | 3 -> tree#remove_node node ; true
       | 2 ->
           tree#update_node ~with_subs:true node { e with name = "/" } ;
           true
       | _ ->
           Log.app (fun m -> m "clicked with button %d on %s" ev.Widget.button e.name);
           true
    )
  in
  (win, tree)

let main () =
  let roots =
    if Array.length Sys.argv > 1 then
      Array.to_list (Array.(sub Sys.argv 1 (length Sys.argv - 1)))
    else
      [Filename.current_dir_name]
  in
  let%lwt roots = Lwt_list.map_p
    (fun name ->
       let%lwt is_dir = is_dir name in
       Lwt.return { name ; is_dir ; selected = false })
      roots
  in
  let (_win, tree) = mk_win () in
  tree#set_roots roots ;
  ignore(tree#grab_focus ()) ;
  Lwt.join [Stk.App.run ()]

let () = Lwt_main.run (main ())